   /*This file is part of KnowMore.org Firefox Extension.

    KnowMore.org Firefox Extension is free software: you can redistribute
    it and/or modify it under the terms of the GNU General Public License 
    as published by the Free Software Foundation, either version 3 of the 
    License, or (at your option) any later version.

    KnowMore.org Firefox Extension is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KnowMore.org Firefox Extension.  If not, see <http://www.gnu.org/licenses/>.

 Contributor(s):
      Ondrej Donek, <ondrejd@gmail.com>
      Dietrich Ayala <autonome@gmail.com>
      E. Cooper <eric@knowmore.org>*/

//namespace
this.KnowMore = this.KnowMore || {}

/**
 * Database service
 *
 * @type {object}
 */
KnowMore.DbService = {};

// Constants with SQL queries
const KM_SQL_CREATE_DB_TABLE1 = 
    "CREATE TABLE [km_profiles] (" +
    "[id] INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL," +
    "[km_id] VARCHAR(255) NULL, " +
    "[km_url] VARCHAR(255) NULL, " + 
    "[cmp_name] VARCHAR(255) NULL, " + 
    "[data_flag] VARCHAR(55) NULL," + 
    "[issue_summary] TEXT NULL," + 
    "[update_time] TIMESTAMP DEFAULT CURRENT_TIMESTAMP NULL, " + 
    "[r_worker] INTEGER NULL, " + 
    "[r_human] INTEGER NULL, " + 
    "[r_politics] INTEGER NULL, " + 
    "[r_enviro] INTEGER NULL, " + 
    "[r_ethics] INTEGER NULL, " + 
    "UNIQUE(km_id, km_url));";
const KM_SQL_CREATE_DB_TABLE2 = 
    "CREATE TABLE [km_urls] (" +
    "[id] INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, " + 
    "[profile_id] INTEGER NOT NULL, " + 
    "[url] VARCHAR(255) NOT NULL, " +
    "[disabled] INTEGER DEFAULT 0," +
    "UNIQUE(profile_id, url));";
const KM_SQL_SELECT_PROFILE_ID_BY_KM_ID =
    "SELECT id FROM [km_profiles] WHERE km_id = ?1";
const KM_SQL_INSERT_PROFILE =
    "INSERT INTO [km_profiles] " +
    "([km_id], " +
    " [km_url], " +
    " [cmp_name], " +
    " [data_flag], " +
    " [issue_summary], " +
    " [update_time], " +
    " [r_worker], " +
    " [r_human], " +
    " [r_politics], " +
    " [r_enviro], " +
    " [r_ethics]) " +
    "VALUES (?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11);";
const KM_SQL_UPDATE_PROFILE =
    "UPDATE [km_profiles] SET " +
    "[km_id] = ?1, " +
    "[km_url] = ?2, " +
    "[cmp_name] = ?3, " +
    "[data_flag] = ?4, " +
    "[issue_summary] = ?5, " +
    "[update_time] = ?6, " +
    "[r_worker] = ?7, " +
    "[r_human] = ?8, " +
    "[r_politics] = ?9, " +
    "[r_enviro] = ?10, " +
    "[r_ethics] = ?11 " +
    "WHERE id = ?12";
const KM_SQL_INSERT_PROFILE_URL =
    "INSERT OR REPLACE INTO [km_urls] ([profile_id], [url]) VALUES (?1, ?2);";
const KM_SQL_SELECT_PROFILE_BY_URL =
    "SELECT t2.* FROM km_urls AS t1 " +
    "LEFT JOIN km_profiles AS t2 ON t2.ID = t1.profile_id " +
    "WHERE t1.url LIKE ?1 AND t1.disabled < (2 - ?2) " +
    "LIMIT 1;";
const KM_SQL_SELECT_PROFILE_BY_RANDOM =
    "SELECT t2.* FROM km_urls AS t1 " +
    "LEFT JOIN km_profiles AS t2 ON t2.ID = t1.profile_id " +
    "ORDER BY RANDOM() " +
    "LIMIT 1;";
const KM_SQL_SELECT_ALL_PROFILES =
    "SELECT t1.* FROM [km_profiles] AS t1 WHERE 1;";
const KM_SQL_SELECT_PROFILE_BY_CMP =
    "SELECT * FROM [km_profiles] WHERE [cmp_name] LIKE ?1 " +
    "LIMIT 1";
const KM_SQL_DISABLE_URL = 
    "UPDATE [km_urls] SET [disabled] = 1 WHERE [url] LIKE ?1 OR [url] LIKE ?2";

const KM_PREF_API_URL =
    "extensions.knowmoreextension.api_url";

const KM_PREF_UPDATE_FREQ =
    "extensions.knowmoreextension.update_frequency";



/**
 * Returns instance of single KnowMore.org profile
 */
KnowMore.DbService.getProfilePrototype = function()
{
  var profile = new Object();
  profile.id            = null;
  profile.km_id         = null;
  profile.km_url        = null;
  profile.cmp_name      = null;
  profile.data_flag     = null;
  profile.issue_summary = null;
  profile.update_time   = null;
  profile.r_worker      = null;
  profile.r_human       = null;
  profile.r_politics    = null;
  profile.r_enviro      = null;
  profile.r_ethics      = null;
  profile.cmp_urls      = [];
  
  /**
   * Add another company's URL
   * @todo Save also URL into database!!!!
   */
  profile.addURL = function(aURL) { profile.cmp_urls.push(aURL); };
  
  /**
   * Returns TRUE whenever given URL match to some company URL
   * @returns {boolean}
   */
  profile.hasURL = function(aURL) { return profile.cmp_urls.indexOf(aURL) != -1; };
  
  return profile;
};


/**
 * Returns instance of single KnowMore.org profile
 */
KnowMore.DbService.getProfileURLPrototype = function()
{
  var profile_url = new Object();
  profile_url.id = null;
  profile_url.profile_id = null;
  profile_url.url = null;
  
  return profile_url;
};


/**
 * ...
 */
KnowMore.DbService.mInitialized = false;


/**
 * ...
 */
KnowMore.DbService.mDbConnect = null;


/**
 * Holds instance of running HTTPRequest
 */
KnowMore.DbService._request = null;


/**
 * Initialize KnowMore database
 *
 * @param aFile {Components.interfaces.nsILocalFile}
 * @returns {boolean}
 */
KnowMore.DbService._construct = function(aFile)
{
  if(!aFile) {
    aFile = KnowMore.Utils.getProfileFile("knowmore.sqlite", ["extensions", "knowmoreextension@knowmore.org"]);
    // Temporarily store the db in profile directory in order to accomodate
    // development shortcuts for no requirement to "install" the extension, see:
    // http://developer.mozilla.org/en/docs/Setting_up_extension_development_environment#Custom_code_location
    // Uncomment this before release:
    //                           ["extensions", "knowmoreextension@knowmore.org"]);
  }
  
  var create_tables = (!aFile.exists()) ? true : false;
  
  if(!KnowMore.DbService.setDbConnection(aFile))
  {
    alert("Unable to open KnowMore database");
    // This is not correct ... here we are in the situation when we haven't
    // cache database but we still have knowmore.org server with its data
    // so we can pull it into the memory and use it as from cache...
    // PS: this error happens only when:
    // - you have no more space on the disk
    // - you haven't write permissions to your profile directory (eg: virus scanner lockdown)
    // - the database is corrupt
  }
  else
  {
    KnowMore.DbService.mDbConnect.executeSimpleSQL("PRAGMA encoding = \"UTF-8\" ");
    
    // We need to create tables and get the initial data from KnowMore.org
    if(create_tables) {
      KnowMore.DbService.createTables();
      KnowMore.DbService.updateData();
    }
    
    // initialize database update timer
    KnowMore.DbService._initTimer();
  }
};
  
  
/**
 * ...
 */
KnowMore.DbService._destruct = function()
{
  KnowMore.DbService.mDbConnect = null;
};

  
/**
 * Initialize timer for updating the database from
 * the web API.
 *
 * @returns {void}
 */
KnowMore.DbService._initTimer = function()
{
  // Setup a cross-session timer to periodically update the database
  var updateManager = Components.classes["@mozilla.org/updates/timer-manager;1"].
                      getService(Components.interfaces.nsIUpdateTimerManager);
  var interval = KnowMore.Utils.getPref(KM_PREF_UPDATE_FREQ, 86400 /* default daily */);
  var updateCallback = {
    _svc: this,
    notify: function(timer) {
      this._svc.updateData();
    }
  };
  updateManager.registerTimer("knowmore-database-update-timer",
                              updateCallback, interval);
};


  /**
   * Get API URL
   * 
   * @returns {string} URL
   */
KnowMore.DbService._getAPIURL = function() {
    return KnowMore.Utils.getPref(KM_PREF_API_URL, null);
};

  
/**
 * Updates cache data
 */
KnowMore.DbService.updateData = function()
{
  var url = KnowMore.DbService._getAPIURL();
  if (!url) {
    url = "http://www.knowmore.org/wiki/api/api.php";
    //Cu.reportError("Unable to get the API URL from preferences");
    // XXX Here should go a proper error message for the end user
    //return;
  }
  // get all profiles considered 'extension-ready'
  url += "?request=profile&flag=e";
  
  // Make request for data
  // XXX Use post!!!
  KnowMore.DbService._request = new XMLHttpRequest();
  KnowMore.DbService._request.open("GET", url, true);
  KnowMore.DbService._request.onreadystatechange = KnowMore.DbService.onDataRequestStateChange;
  KnowMore.DbService._request.send(null);
};

  
/**
 * Fired when HTTPRequest for data change its status
 *
 * @param aEvent {nsIDOMEvent}
 */
KnowMore.DbService.onDataRequestStateChange = function(aEvent)
{
if(KnowMore.DbService._request.readyState == 4)
  {
    if(KnowMore.DbService._request.status == 200)
    {
      var json_str = KnowMore.DbService._request.responseText;
      
      
      var data_obj;
      try {
        // Fx3
        if (Components.interfaces.nsIJSON) {
          var json = Components.classes["@mozilla.org/dom/json;1"].
              createInstance(Components.interfaces.nsIJSON);
          data_obj = json.decode(json_str);
        }
        // Fx2
        else
          data_obj = JSON.parse(json_str);
      } catch(ex) {
        Components.utils.reportError("Unable to decode response: " + ex);
      }
      
      if(data_obj.profiles)
      {
        // XXX wrap the update in a transaction for performance

        try {
          for(var i=0; i<data_obj.profiles.length; i++)
          {
            var profile = data_obj.profiles[i];
            var cmp = KnowMore.DbService.getProfilePrototype();
            cmp.km_id = profile.id;
            cmp.km_url = profile.km_url;
            cmp.cmp_name = profile.company_name;
            cmp.data_flag = profile.data_flag;
            cmp.issue_summary = profile.issue_summary;
            cmp.update_time = profile.update_time;
            cmp.r_worker = profile.r_worker;
            cmp.r_human = profile.r_human;
            cmp.r_politics = profile.r_politics;
            cmp.r_enviro = profile.r_enviro;
            cmp.r_ethics = profile.r_ethics;
            
            if(profile.company_url) 
              for(var j=0; j<profile.company_url.length; j++) 
                cmp.addURL(profile.company_url[j]);
            
            KnowMore.log("adding profile for " + cmp.cmp_name + ": " + profile.company_url.length + " -> " + cmp.cmp_urls.length);
            KnowMore.DbService.addProfile(cmp);
          }
        } catch(ex) {
          KnowMore.log("error: " + ex);
          Components.utils.reportError("error during profile import: " + ex);
          return;
        }
      }
    }
  }
};
  
  
/**
 * Set up connection to SQLite database
 *
 * @param aFile {Components.interfaces.nsILocalFile}
 * @returns {boolean}
 */
KnowMore.DbService.setDbConnection = function(aFile)
{
  var dbSrv = Components.classes["@mozilla.org/storage/service;1"].
      getService(Components.interfaces.mozIStorageService);
  
  KnowMore.DbService.mDbConnect = dbSrv.openDatabase(aFile);
  
  var counter = 0;
  while(!KnowMore.DbService.mDbConnect.connectionReady) 
    if(counter++ == 1000000) return false;
  
  return true;
};
  
  
KnowMore.DbService.startTransaction = function()
{
  //KnowMore.DbService.mDbConnect.beginTransaction();
};


KnowMore.DbService.finishTransaction = function()
{
  //KnowMore.DbService.mDbConnect.commitTransaction();
};


KnowMore.DbService.rollbackTransaction = function()
{
  //KnowMore.DbService.mDbConnect.rollbackTransaction();
};


/**
 * Removes all tables of the database
 *
 * @returns {boolean}
 */
KnowMore.DbService.removeTables = function()
{ 
  try {
    KnowMore.DbService.startTransaction();
    
    KnowMore.DbService.mDbConnect.executeSimpleSQL("DROP TABLE [km_profiles] ");
    KnowMore.DbService.mDbConnect.executeSimpleSQL("DROP TABLE [km_urls] ");
    
    KnowMore.DbService.finishTransaction();
    
    return true;
  } catch(e) {
    Components.utils.reportError(e);
    return false;
  }
};
  
  
/**
 * Creates KnowMore database
 *
 * @returns {boolean}
 */
KnowMore.DbService.createTables = function()
{ 
  try {
    KnowMore.DbService.startTransaction();
    KnowMore.DbService.mDbConnect.executeSimpleSQL(KM_SQL_CREATE_DB_TABLE1);
    KnowMore.DbService.mDbConnect.executeSimpleSQL(KM_SQL_CREATE_DB_TABLE2);
    KnowMore.DbService.finishTransaction();
    
    return true;
  } catch(e) {
    Components.utils.reportError(e);
    return false;
  }
};


// ========================================================================
// Profiles

/**
 * Add profile to the database. Returns ID of inserted item.
 * 
 * @param aProfile {object}
 * @returns {integer}
 */
KnowMore.DbService.addProfile = function(aProfile) 
{
  // XXX Use a transaction

  var getIdStmt =
    KnowMore.DbService.mDbConnect.createStatement(KM_SQL_SELECT_PROFILE_ID_BY_KM_ID);
  getIdStmt.bindInt32Parameter(0, aProfile.km_id);
  var profileStmt, profileId;
  if (getIdStmt.executeStep()) {
    profileId = getIdStmt.getInt32(0);
    profileStmt = KnowMore.DbService.mDbConnect.createStatement(KM_SQL_UPDATE_PROFILE);
    profileStmt.bindInt32Parameter(11, profileId);
  }
  else
    profileStmt = KnowMore.DbService.mDbConnect.createStatement(KM_SQL_INSERT_PROFILE);

  profileStmt.bindInt32Parameter(0, aProfile.km_id);
  profileStmt.bindStringParameter(1, aProfile.km_url);
  profileStmt.bindUTF8StringParameter(2, aProfile.cmp_name);
  profileStmt.bindStringParameter(3, aProfile.data_flag);
  profileStmt.bindUTF8StringParameter(4, aProfile.issue_summary);
  profileStmt.bindStringParameter(5, aProfile.update_time);
  profileStmt.bindInt32Parameter(6, aProfile.r_worker);
  profileStmt.bindInt32Parameter(7, aProfile.r_human);
  profileStmt.bindInt32Parameter(8, aProfile.r_politics);
  profileStmt.bindInt32Parameter(9, aProfile.r_enviro);
  profileStmt.bindInt32Parameter(10, aProfile.r_ethics);
  profileStmt.execute();

  if (!profileId)
    profileId = KnowMore.DbService.mDbConnect.lastInsertRowID;

  for(var i=0; i < aProfile.cmp_urls.length; i++) {
    var urlStmt = KnowMore.DbService.mDbConnect.createStatement(KM_SQL_INSERT_PROFILE_URL);
    urlStmt.bindInt32Parameter(0, profileId);
    urlStmt.bindStringParameter(1, aProfile.cmp_urls[i]);
    urlStmt.execute();
  }
  
  return KnowMore.DbService.mDbConnect.lastInsertRowID;
};
  
  
/**
 * Gets single profile with given id from the database
 * 
 * @param aId {integer}
 * @returns {KM_ProfileDatatypePrototype} Or NULL
 */
KnowMore.DbService.getProfile = function(aId)  
{
  // ...
};


/**
 * Returns array with all profiles that are stored in the database.
 * 
 * @return {array} Returns array with instances of KM_ProfileDatatypePrototype.
 */
KnowMore.DbService.getAllProfiles = function()
{
  var profiles = new Array();
  var stmt = KnowMore.DbService.mDbConnect.createStatement(KM_SQL_SELECT_ALL_PROFILES);
  
  while(stmt.executeStep())
    profiles.push(KnowMore.DbService.bindStatementToObject(stmt));
  
  return profiles;
};


/**
 * Returns total count of profiles
 *
 * @returns {integer}
 */
KnowMore.DbService.getProfilesCount = function()
{
  var stmt = KnowMore.DbService.mDbConnect.createStatement("SELECT count(ID) FROM [km_profiles] WHERE 1 ");
  
  while(stmt.executeStep())
    return stmt.getInt32(0);
  
  return null;
};


/**
 * Returns profile by given URL
 *
 * @param aUrl {string}
 * @returns {object}
 */
KnowMore.DbService.getProfileByURL = function(aUrl, disable)
{
  if(aUrl != null || aUrl != "") {

    var stmt = KnowMore.DbService.mDbConnect.createStatement(KM_SQL_SELECT_PROFILE_BY_URL);

    if(!disable) 
        disable = 0;

    stmt.bindStringParameter(0, "http%://%." + aUrl + "%");
    stmt.bindStringParameter(1, disable);
    
    if(stmt.executeStep()) 
      return KnowMore.DbService.bindStatementToObject(stmt);
        
  }
  
  return null;
};


/**
 * Returns profile by a given company name
 *
 * @param aCmp {string}
 * @returns {object}
 */

KnowMore.DbService.getProfileByCmp = function(aCmp)
{
    if (aCmp != null || aCmp != "") {
        var stmt = KnowMore.DbService.mDbConnect.createStatement(KM_SQL_SELECT_PROFILE_BY_CMP);
        stmt.bindStringParameter(0, "%" + aCmp + "%");
        
        if (stmt.executeStep())
            return KnowMore.DbService.bindStatementToObject(stmt);
    }

    return null;
}

/**
 * Returns a random profile
 *
 * @returns {object}
 */
KnowMore.DbService.getRandomProfile = function()
{
  var stmt = KnowMore.DbService.mDbConnect.createStatement(KM_SQL_SELECT_PROFILE_BY_RANDOM);

  if(stmt.executeStep())
    return KnowMore.DbService.bindStatementToObject(stmt);
  
  return null;
};

/**
 * Disables a url in [km_urls] 
 *
 * @param aUrl {string}
 * @returns {object}
 */

KnowMore.DbService.disableProfileUrl = function (aUrl)
{
    var stmt = KnowMore.DbService.mDbConnect.createStatement(KM_SQL_DISABLE_URL);
    stmt.bindStringParameter(0, "http%://%." + aUrl + "%");

    if(stmt.executeStep()) {
        return true;}

    return false;
}

/**
 * Converts a statement record into a profile object.
 *
 * @returns {object}
 */
KnowMore.DbService.bindStatementToObject = function(stmt)
{
  var profile = KnowMore.DbService.getProfilePrototype();
  profile.id            = stmt.getInt32(0);
  profile.km_id         = stmt.getString(1);
  profile.km_url        = stmt.getString(2);
  profile.cmp_name      = stmt.getUTF8String(3);
  profile.data_flag     = stmt.getString(4);
  profile.issue_summary = stmt.getUTF8String(5);
  // update_time
  profile.r_worker      = stmt.getInt32(7);
  profile.r_human       = stmt.getInt32(8);
  profile.r_politics    = stmt.getInt32(9);
  profile.r_enviro      = stmt.getInt32(10);
  profile.r_ethics      = stmt.getInt32(11);
  
  return profile;
}; 
